# -*- coding: utf-8 -*-

# Standard Library
import copy
import json

# Vendor Library
import objectpack

from m3.actions import ACD, ControllerCache, OperationResult, PreJsonResult
from m3_ext.ui import all_components as ext
from m3_ext.ui.icons import Icons
from m3.actions.exceptions import ApplicationLogicException

# Project Library
from kinder.core.declaration_status.enum import DSS
from kinder.core.declaration_status.models import DeclarationStatus
from kinder.core.helpers import get_report_filepath
from concentrator.models import ChangeStatus, ChangeDeclaration
from concentrator.rules import DeclarationStatusChangeRule
from concentrator.changes.reports.report import ReportChanges
from concentrator.webservice.service import (
    send_update_application_state)
from concentrator.change import StorageHelper
from ..models import ChangeSource


# TODO: Обработка ошибок!
class ChangesPack(objectpack.ObjectPack):
    """
    Пак для вкладки "Изменения с ЕПГУ"
    """

    title = u"Изменения с ЕПГУ"
    __PERMISSION_NS = "kinder.plugins.concentrator.changes.actions.ChangesPack"
    need_check_permission = True

    columns = [
        {
            'data_index': 'create',
            'header': u'Дата',
            'searchable': False,
            'sortable': True,
        },
        {
            'data_index': 'fields',
            'header': u'Поля',
            'searchable': False,
            'sortable': False,
        },
        {
            'data_index': 'old_values',
            'header': u'Старое значение',
            'searchable': False,
            'sortable': False,
        },
        {
            'data_index': 'new_values',
            'header': u'Новое значение',
            'searchable': False,
            'sortable': False,
        },
        {
            'data_index': 'state',
            'header': u'Статус',
            'searchable': False,
            'sortable': True,
        },
        {
            'data_index': 'user',
            'header': u'Пользователь',
            'searchable': False,
            'sortable': True,
        },
        {
            'data_index': 'commentary',
            'header': u'Комментарий',
            'searchable': False,
            'sortable': True,
        },
    ]

    def __init__(self):
        super(ChangesPack, self).__init__()

        self.changes_print_action = ChangesPrintAction()
        self.changes_rows_action = ChangesRowsAction()

        self.replace_action('rows_action', self.changes_rows_action)

        self.actions.extend([
            self.changes_print_action,
        ])


class ChangesRowsAction(objectpack.ObjectRowsAction):
    """
    Экшн преобразует набор изменений для грида во вкладке "Изменения с ЕПГУ"
    """

    def context_declaration(self):
        return [ACD(name='declaration_id', required=True, type=int)]

    def run(self, request, context):
        new_self = copy.copy(self)
        new_self.request = request
        new_self.context = context

        new_self.query = ChangeDeclaration.objects.filter(
            declaration_id=context.declaration_id).order_by('create')

        total = new_self.get_total_count()

        new_self.apply_sort_order()
        new_self.apply_limit()

        rows = []

        for row in new_self.query:
            change_set = {
                'fields': [],
                'old_values': [],
                'new_values': []
            }
            changes_rows = StorageHelper.get_change(row)
            for changes_row in changes_rows:
                change_set['fields'].append(
                    changes_row['field'])
                change_set['old_values'].append(
                    unicode(changes_row['old_value']))
                change_set['new_values'].append(
                    unicode(changes_row['new_value']))

            rows.append({
                'id': row.id,
                'create': row.create,
                'fields': u', '.join(change_set['fields']),
                'old_values': u', '.join(change_set['old_values']),
                'new_values': u', '.join(change_set['new_values']),
                'state': ChangeStatus.values[row.state],
                'user': row.user.get_full_name().title() if row.user else '',
                'commentary': row.commentary
            })

        return PreJsonResult({'rows': rows, 'total': total})


class ChangesPrintAction(objectpack.BaseAction):
    """
    Экшн отдает .xls файл с пустым/заполненным заявлением на изменение данных
    """
    __PERMISSION_NS = "kinder.plugins.concentrator.changes.actions.ChangesPrintAction"
    verbose_name = u"Печать"
    need_check_permission = True

    def context_declaration(self):
        return [ACD(name='blank', required=True, type=bool),
                ACD(name='id', required=False, type=int)]

    def run(self, request, context):
        templates = {
            True: 'changes_blank.xls',
            False: 'changes.xls'
        }

        template = templates[context.blank]
        (_, xls_file_abs, url) = get_report_filepath('xls')
        rep = ReportChanges(template, xls_file_abs, request, context)
        rep.make_report()

        return OperationResult(code="function() {location.href='%s';}" % url)


class ChangesDetailPack(objectpack.ObjectPack):
    """
    Пак детального просмотра набора изменений
    """
    __PERMISSION_NS = "kinder.plugins.concentrator.changes.actions.ChangesDetailPack"
    title = u"Изменения с ЕПГУ"

    need_check_permission = True

    allow_paging = False

    columns = [
        {
            'data_index': 'field',
            'header': u'Поле',
            'searchable': False,
            'sortable': False,
        },
        {
            'data_index': 'old_value',
            'header': u'Старое значение',
            'searchable': False,
            'sortable': False,
        },
        {
            'data_index': 'new_value',
            'header': u'Новое значение',
            'searchable': False,
            'sortable': False,
        },
    ]

    def __init__(self):
        super(ChangesDetailPack, self).__init__()

        self.replace_action('list_window_action', ChangesDetailWindowAction())
        self.changes_apply_action = ChangesApplyAction()
        self.changes_reject_action = ChangesRejectAction()
        self.changes_rows_action = ChangesDetailRowsAction()

        self.replace_action('rows_action', self.changes_rows_action)

        self.actions.extend([
            self.changes_apply_action,
            self.changes_reject_action
        ])
        self.sub_permissions = super(
            ChangesDetailPack, self
        ).sub_permissions.copy()
        self.sub_permissions[self.list_window_action.perm_code] = \
            self.list_window_action.verbose_name


class ChangesDetailWindowAction(objectpack.ObjectListWindowAction):
    """
    Экшн отдает окно с детализацией изменений
    """
    __PERMISSION_NS = "kinder.plugins.concentrator.changes.actions.ChangesDetailWindowAction"
    verbose_name = u"Подтверждение"
    need_check_permission = True

    def context_declaration(self):
        return [ACD(name='id', required=True, type=int)]

    def configure_window(self):
        self.win.template_globals = 'ui-js/changes_detail.js'
        self.win.grid_id = self.win.grid.client_id
        self.win.grid.top_bar.hidden = True
        self.win.modal = True

        change = ChangeDeclaration.objects.get(id=self.context.id)
        if change.state in [ChangeStatus.WAIT, ChangeStatus.ACCEPT]:

            self.win.print_button = ext.ExtButton(
                text=u"Распечатать заявление",
                handler='printChanges',
                icon_cls=Icons.PRINTER
            )

            self.win.buttons.insert(0, self.win.print_button)

            self.win.changes_print_url = ControllerCache.find_pack(
                'concentrator.changes.actions.'
                'ChangesPack').changes_print_action.get_absolute_url()

        if change.state == ChangeStatus.WAIT:

            self.win.apply_button = ext.ExtButton(
                text=u"Применить изменения",
                handler='applyChanges',
                icon_cls=Icons.ACCEPT
            )
            self.win.buttons.insert(0, self.win.apply_button)

            self.win.reject_button = ext.ExtButton(
                text=u"Отказать",
                handler='rejectChanges',
                icon_cls=Icons.M3_DELETE
            )
            self.win.buttons.insert(0, self.win.reject_button)

            self.win.changes_apply_url = ControllerCache.find_pack(
                'concentrator.changes.actions.'
                'ChangesDetailPack').changes_apply_action.get_absolute_url()
            self.win.changes_reject_url = ControllerCache.find_pack(
                'concentrator.changes.actions.'
                'ChangesDetailPack').changes_reject_action.get_absolute_url()


class ChangesDetailRowsAction(objectpack.ObjectRowsAction):
    """
    Экшн преобразует набор изменений для грида с детализацией изменений
    """

    def context_declaration(self):
        return [ACD(name='id', required=True, type=int)]

    def run(self, request, context):
        changes = ChangeDeclaration.objects.get(
            id=context.id)

        rows = StorageHelper.get_change(changes)
        total = len(rows)
        return PreJsonResult({'rows': rows, 'total': total})


class ChangesApplyAction(objectpack.BaseAction):
    """
    Экшн применяет набор изменений и возвращает обновленные поля
    """

    # Если значение в поле нельзя сменить стандартным способом,
    # то используем для этого кастомные правила.
    # Кастомное правило это класс с classmethod-ом change,
    # в который передается обьект и новое значение поля
    CHANGE_RULE_MAP = {
        ('Declaration', 'status'): DeclarationStatusChangeRule
    }

    def context_declaration(self):
        return [ACD(name='id', required=True, type=int),
                ACD(name='comment', required=True, type=unicode,
                    default=u'')]

    def run(self, request, context):
        change = ChangeDeclaration.objects.get(
            id=context.id)

        try:
            updated_fields = StorageHelper.apply_changes(
                change, request,
                context.comment)
        except ApplicationLogicException as e:
            change.commentary = u"Невозможно применить: {message}".format(
                message=e.message)
            change.save()
            return OperationResult.by_message(e.message)

        # При принятии данных поступивших через NewApplication
        if change.source == ChangeSource.NEW_APPLICATION:
            doc_confirmating_status = DeclarationStatus.objects.get(
                code=DSS.DUL_CONFIRMATING)
            change.declaration.change_status(
                doc_confirmating_status,
                why_change=u"Авто-смена статуса заявления после "
                           u"поступления новой заявки с Концентратора")
            change.declaration.date = change.create.date()
            change.declaration.save()

        return PreJsonResult(data={'updated_fields': updated_fields})


class ChangesRejectAction(objectpack.BaseAction):
    """
    Экшн отклоняет набор изменений
    """

    def context_declaration(self):
        return [ACD(name='id', required=True, type=int),
                ACD(name='comment', required=True, type=unicode,
                    default=u'')]

    def _update_application_state(self, declaration):
        """
        Отправка информации о заявке.
        """

        send_update_application_state(declaration)

    def _reject_status_change(self, changes):
        """
        Имеется ли в наборе изменений изменения статуса заявки
        """

        changes_data = json.loads(
            changes.data)

        if 'Declaration' in changes_data:
            for change in changes_data['Declaration']:
                if 'status' in change:
                    return True

        return False

    def run(self, request, context):
        changes = ChangeDeclaration.objects.get(
            id=context.id)
        user = getattr(request, 'user', None)
        changes.user = user
        changes.state = ChangeStatus.REJECT
        changes.commentary = context.comment
        changes.save()

        # При отмене данных поступивших через UpdateApplication
        if changes.source == ChangeSource.NEW_APPLICATION:
            refused_status = DeclarationStatus.objects.get(
                code=DSS.REFUSED)
            changes.declaration.change_status(
                refused_status,
                why_change=u"Отмена принятия данных поступивших с ЕПГУ"
            )
        elif self._reject_status_change(changes):
            self._update_application_state(changes.declaration)

        return OperationResult()
